import numpy.ma as ma
import numpy as np
import pandas as pd
from typing import List

from genome_tools.genomic_interval import GenomicInterval, genomic_intervals_to_df, df_to_genomic_intervals

from hotspot3.io.logging import WithLoggerAndInterval
from hotspot3.background_fit.fit import GlobalBackgroundFit, StridedBackgroundFit, WindowBackgroundFit
from hotspot3.helpers.models import FitResults, WindowedFitResults, NotEnoughDataForContig
from hotspot3.helpers.format_converters import set_series_row_to_df, convert_fit_results_to_series, fit_stats_df_to_fallback_fit_results
from hotspot3.helpers.stats import check_valid_nb_params, threhold_from_bg_tag_proportion
from hotspot3.helpers.utils import interpolate_nan


class SegmentalFit(WithLoggerAndInterval):
    def filter_signal_to_segment(self, agg_cutcounts: ma.MaskedArray, segment: GenomicInterval=None) -> ma.MaskedArray:
        if segment is None:
            segment = self.genomic_interval
        return agg_cutcounts[segment.start:segment.end]
    
    def per_bp_background_model_fit(
            self,
            agg_cutcounts: ma.MaskedArray,
            segments: pd.DataFrame,
        ):
        per_window_trs = np.full(agg_cutcounts.shape[0], np.nan, dtype=np.float16)
        windowed_fit_results = WindowedFitResults(
            p=per_window_trs.copy(),
            r=per_window_trs.copy(),
            enough_bg_mask=np.zeros(agg_cutcounts.shape[0], dtype=bool)
        )
        segments_genomic_intervals = df_to_genomic_intervals(segments)

        for i, segment_interval in enumerate(segments_genomic_intervals):
            start = int(segment_interval.start)
            end = int(segment_interval.end)
            segment_fit_result = fit_stats_df_to_fallback_fit_results(segments.iloc[[i]])
            if not check_valid_nb_params(segment_fit_result):
                continue
            signal_at_segment = self.filter_signal_to_segment(agg_cutcounts, segment_interval)
            fit_res = self.fit_segment_params(
                signal_at_segment,
                segment_fit_result.fit_threshold,
                fallback_fit_results=segment_fit_result,
                genomic_interval=segment_interval
            )
            windowed_fit_results.r[start:end] = fit_res.r
            windowed_fit_results.p[start:end] = fit_res.p
            windowed_fit_results.enough_bg_mask[start:end] = fit_res.enough_bg_mask
            per_window_trs[start:end] = segment_fit_result.fit_threshold
        return windowed_fit_results, per_window_trs
    
    def fit_per_segment_bg_model(
            self,
            agg_cutcounts: ma.MaskedArray,
            bad_segments: List[GenomicInterval],
            fallback_fit_results: FitResults,
            min_bg_tag_proportion: np.ndarray=None
        ):
        total_len = agg_cutcounts.count()
        intervals_stats = genomic_intervals_to_df(bad_segments).drop(columns=['chrom', 'name'])
        if min_bg_tag_proportion is not None:
            assert len(min_bg_tag_proportion) == len(bad_segments), "min_bg_tag_proportion should have the same length as segments"
        self.logger.debug(f"{self.genomic_interval.chrom}: Fitting background model for {len(bad_segments)} segments")
        for i, segment_interval in enumerate(bad_segments):
            signal_at_segment = self.filter_signal_to_segment(agg_cutcounts, segment_interval)

            g_fit = self.copy_with_params(
                GlobalBackgroundFit,
                name=segment_interval.to_ucsc()
            )
            reached_max_q = False
            try:
                step = self.get_optimal_segment_step(total_len, len(segment_interval))
                if min_bg_tag_proportion is not None:
                    valid_count = threhold_from_bg_tag_proportion(
                        signal_at_segment.compressed(),
                        min_bg_tag_proportion[i],
                    )
                    min_bg_quantile = g_fit.get_bg_quantile_from_tr(
                        signal_at_segment[::step].filled(np.nan), # calc quantile on the same data that goes in g_fit
                        valid_count
                    )
                    if min_bg_quantile > g_fit.config.max_background_prop:
                        reached_max_q = True
                    g_fit.config.min_background_prop = min(
                        g_fit.config.max_background_prop,
                        min_bg_quantile
                    )

                segment_fit_results = g_fit.fit(
                    signal_at_segment,
                    step=step,
                    fallback_fit_results=fallback_fit_results
                )
                success_fit = True

            except NotEnoughDataForContig:
                segment_fit_results = FitResults()
                success_fit = False
            
            fit_series = convert_fit_results_to_series(
                segment_fit_results,
                fit_type='segment',
                success_fit=success_fit,
                reached_max_q=reached_max_q
            )
            set_series_row_to_df(intervals_stats, fit_series, i)
        return intervals_stats


    def add_fallback_fit_stats(
            self,
            fallback_fit_results: FitResults,
            segment_stats: pd.DataFrame
        ):
        intervals_stats = genomic_intervals_to_df([self.genomic_interval]).drop(
            columns=['chrom', 'name']
        )
        intervals_stats['BAD'] = 0
        fit_series = convert_fit_results_to_series(
            fallback_fit_results,
            fit_type='global',
            success_fit=True,
        )
        set_series_row_to_df(intervals_stats, fit_series, 0)
        for df in [intervals_stats, segment_stats]:
            df[['bg_r', 'bg_p', 'rmsea', 'bg_tr']] = df[['bg_r', 'bg_p', 'rmsea', 'bg_tr']].astype(np.float32)
        return pd.concat([intervals_stats, segment_stats], ignore_index=True)


    def get_optimal_segment_step(self, total_len, segment_len):
        # Scale step according to the length of segment
        return max(
            20, 
            round(self.config.chromosome_fit_step / total_len * segment_len)
        )

    def fit_segment_params(
            self,
            signal_at_segment: ma.MaskedArray,
            thresholds: np.ndarray,
            fallback_fit_results: FitResults,
            genomic_interval: GenomicInterval=None
        ) -> WindowedFitResults:
        if genomic_interval is None:
            genomic_interval = self.genomic_interval

        w_fit = self.copy_with_params(
            WindowBackgroundFit,
            name=genomic_interval.to_ucsc()
        )

        fit_res = w_fit.fit(signal_at_segment, per_window_trs=thresholds)
        success_fits = check_valid_nb_params(fit_res) & fit_res.enough_bg_mask

        need_global_fit = ~success_fits & fit_res.enough_bg_mask
        if np.any(need_global_fit):
            fit_res.r[need_global_fit] = fallback_fit_results.r

            fit_res.p[need_global_fit] = w_fit.fit(
                signal_at_segment,
                per_window_trs=thresholds,
                fallback_fit_results=fallback_fit_results
            ).p[need_global_fit]
        self.logger.debug(f"{genomic_interval.to_ucsc()}: Successfully fit background model for {np.sum(success_fits):,} bp. Use segment r for {np.sum(need_global_fit):,} bp")
        return fit_res


class ChromosomeFit(WithLoggerAndInterval):
    
    def fit_segment_thresholds(
            self,
            agg_cutcounts: ma.MaskedArray,
            step=None
        ):
        if step is None:
            step = self.config.chromosome_fit_step

        g_fit = self.copy_with_params(GlobalBackgroundFit)
        global_fit_results = g_fit.fit(agg_cutcounts, step=step)
        if not check_valid_nb_params(global_fit_results):
            raise NotEnoughDataForContig
        strided_fit = self.copy_with_params(StridedBackgroundFit)
        fit_threshold = strided_fit.find_thresholds_at_chrom_quantile(
            agg_cutcounts,
            global_fit_results.fit_quantile,
            bg_window=self.config.bg_window
        )
        fit_threshold2 = strided_fit.find_thresholds_at_chrom_quantile(
            agg_cutcounts,
            global_fit_results.fit_quantile,
            bg_window=self.config.bg_window_small
        )
        fit_threshold = interpolate_nan(np.minimum(fit_threshold, fit_threshold2))
        self.logger.debug(f"{self.name}: Signal thresholds approximated")
        return fit_threshold, global_fit_results



import os
import shutil
import pandas as pd
import tempfile
import pyBigWig
import numpy as np
import subprocess

from genome_tools.helpers import df_to_tabix
from hotspot3.helpers.models import ProcessorOutputData, NotEnoughDataForContig

from hotspot3.io import parallel_write_partitioned_parquet
from hotspot3.io.logging import WithLoggerAndInterval, WithLogger
from hotspot3.signal_smoothing import normalize_density


class ChromWriter(WithLoggerAndInterval):

    def parallel_write_chromdata_to_parquet(self, data_df, path, chrom_names, compression_level=22):
        """
        Workaround for writing parquet files for chromosomes in parallel.
        """
        chrom_name = self.genomic_interval.chrom
        if data_df is None:
            raise NotEnoughDataForContig
        if isinstance(data_df, ProcessorOutputData):
            data_df = data_df.data_df
        data_df['chrom'] = pd.Categorical(
            [chrom_name] * data_df.shape[0],
            categories=chrom_names
        )
        parallel_write_partitioned_parquet(
            data_df,
            field_names=[chrom_name],
            partition_cols=['chrom'],
            path=path,
            tmp_dir=self.config.tmp_dir,
            compression_level=compression_level
        )


class GenomeWriter(WithLogger):

    @staticmethod
    def write_bigwig(df: pd.DataFrame, col: str, outpath, chrom_sizes: dict):
        with pyBigWig.open(outpath, 'w') as bw:
            bw.addHeader(list(chrom_sizes.items()))
            chroms = df['chrom'].to_list()
            starts = df['start'].to_list()
            ends = df['end'].to_list()
            values = df[col].to_list()
            bw.addEntries(chroms, starts, ends=ends, values=values)

    def df_to_bigwig(self, df: pd.DataFrame, outpath: str, chrom_sizes: dict, col='value'):
        try:
            self.write_bigwig(df, col, outpath, chrom_sizes)
        except RuntimeError as e:
            self.logger.warning(
                "pyBigWig.addEntries failed: Entries are out of order.\n"
                "Ensure that:\n"
                "  1. Chromosomes in chrom_sizes file follow UCSC order (chr1, chr2, ..., chrX, chrY, chrM).\n"
                "  2. Non-standard contigs (e.g., chrUn_*, GL*, KI*) are filtered or placed at the end.\n"
                f"Attempting to filter non-standard contigs and retrying.\n"
            )
            non_standard_chroms = [chrom for chrom in chrom_sizes if '_' in chrom]
            if non_standard_chroms:
                self.logger.info(
                    f"Filtering out non-standard chromosomes: {', '.join(non_standard_chroms)}"
                )
                df = df[~df['chrom'].isin(non_standard_chroms)]
                try:
                    self.write_bigwig(df, col, outpath, chrom_sizes)
                except RuntimeError as e:
                    self.logger.error(f"Retrying failed: {e}")
                    raise RuntimeError(
                        "pyBigWig.addEntries failed: Entries are still out of order.\n"
                        "Ensure that:\n"
                        "  1. Chromosomes in chrom_sizes file follow UCSC order (chr1, chr2, ..., chrX, chrY, chrM).\n"
                        "  2. Non-standard contigs (e.g., chrUn_*, GL*, KI*) are filtered or placed at the end.\n"
                        f"Original error: {e}"
                    )


    def sanitize_path(self, path):
        """
        Call to properly clean up the path to replace parquets
        """
        if os.path.exists(path):
            if os.path.isdir(path):
                shutil.rmtree(path)
            else:
                os.remove(path)
    
    def df_to_tabix(self, df, outpath):
        df_to_tabix(df, outpath)
    

    def thresholds_df_to_bw(self, thresholds: pd.DataFrame, save_path, total_cutcounts, chrom_sizes):
        thresholds['end'] = thresholds['start'] + self.config.bg_track_step
        thresholds['tr'] = normalize_density(thresholds['tr'], total_cutcounts)
        self.df_to_bigwig(
            thresholds,
            save_path,
            chrom_sizes=chrom_sizes,
            col='tr'
        )

    def density_to_bw(self, density_data: pd.DataFrame, save_path, chrom_sizes):
        density_data['end'] = density_data['start'] + self.config.density_track_step
        self.logger.debug(f"Converting density to bigwig")
        self.df_to_bigwig(
            density_data,
            save_path,
            chrom_sizes=chrom_sizes,
            col='normalized_density'
        )

    def fit_stats_to_bw(
            self,
            fit_stats: pd.DataFrame,
            outpath_bw,
            total_cutcounts,
            chrom_sizes
        ):
        fit_stats = fit_stats.query('fit_type == "segment"')[
            ['chrom', 'start', 'end', 'background']
        ]
        fit_stats['background'] = normalize_density(
            fit_stats['background'],
            total_cutcounts
        )

        self.df_to_bigwig(
            fit_stats,
            outpath_bw,
            chrom_sizes=chrom_sizes,
            col='background'
        )
    
    def save_cutcounts(self, total_cutcounts, total_cutcounts_path):
        np.savetxt(total_cutcounts_path, [total_cutcounts], fmt='%d')
    
    def df_to_bigbed(self, df: pd.DataFrame, chrom_sizes, outpath):
        with tempfile.NamedTemporaryFile(suffix=".bed") as temp_sorted_bed:
            df.to_csv(temp_sorted_bed.name, sep='\t', header=False, index=False)
            try:
                subprocess.run(
                    ["bedToBigBed", temp_sorted_bed.name, chrom_sizes, outpath],
                    capture_output=True,
                    text=True, 
                    check=True
                )
            except subprocess.CalledProcessError as e:
                self.logger.warning(f"Error converting to BigBed: {e}, {e.stderr}")

    def merge_partitioned_parquets(self, parquet_old, parquet_new):
        for file in os.listdir(parquet_old):
            new_path = os.path.join(parquet_new, file)
            if not os.path.exists(new_path):
                shutil.move(os.path.join(parquet_old, file), new_path)
        
        if self.config.save_debug:
            shutil.move(parquet_old, f"{parquet_old}.iter1")
        else:
            shutil.rmtree(parquet_old)
        shutil.move(parquet_new, parquet_old)


from pathlib import Path
import networkx as nx

def fallback_path():
    def decorator(func):
        name = func.__name__
        private_name = f"_{name}"

        @property
        def wrapper(self):
            override = getattr(self, private_name, None)

            if override is not None:
                path = Path(override)
                if not path.exists():
                    raise FileNotFoundError(f"Provided path for `{name}` does not exist: {path}")
                result = path
            else:
                result = func(self)
                if result is None:
                    raise ValueError(f"Path for `{name}` is not set")
            return str(result)

        @wrapper.setter
        def wrapper(self, value):
            setattr(self, private_name, value)

        return wrapper
    return decorator


class Hotspot3Paths:
    def __init__(
            self,
            outdir,
            sample_id,
            save_density=False,
            **kwargs
        ):
        self.outdir = outdir
        self.sample_id = sample_id
        self.outdir = Path(self.outdir)

        allowed_keys = {
            k for k, v in self.__class__.__dict__.items()
        }
        unknown_keys = set(kwargs.keys()) - allowed_keys
        if len(unknown_keys) > 0:
            raise ValueError(f"Unknown keys provided: {unknown_keys}")
        for key, value in kwargs.items():
            self.__setattr__(f"_{key}", value)
            if key == "cutcounts" and value is not None:
                self._total_cutcounts = self.total_cutcounts
        
        self.solver = StepsSolver(self, save_density)
    
    def was_set(self, prop):
        return getattr(self, f"_{prop}", None) is not None

    @property
    def debug_dir(self):
        path = self.outdir / "debug"
        path.mkdir(parents=True, exist_ok=True)
        return path

    @fallback_path()
    def bam(self):
        return None

    @fallback_path()
    def cutcounts(self):
        return self.outdir / f"{self.sample_id}.cutcounts.bed.gz"

    @property
    def total_cutcounts(self):
        if self.cutcounts is None:
            return None
        return self.cutcounts.replace('.cutcounts.bed.gz', '.total_cutcounts')

    @fallback_path()
    def smoothed_signal(self):
        return self.debug_dir / f"{self.sample_id}.smoothed_signal.parquet"

    @fallback_path()
    def fit_params(self):
        return self.debug_dir / f"{self.sample_id}.fit_params.parquet"

    @fallback_path()
    def pvals(self):
        return self.outdir / f"{self.sample_id}.pvals.parquet"

    @fallback_path()
    def fdrs(self):
        return self.debug_dir / f"{self.sample_id}.fdrs.parquet"

    @fallback_path()
    def normalized_density(self):
        return self.outdir / f"{self.sample_id}.normalized_density.bw"
    
    @fallback_path()
    def per_region_background(self):
        return self.debug_dir / f"{self.sample_id}.per_segment.background.bw"
    
    @fallback_path()
    def per_region_stats(self):
        return self.outdir / f"{self.sample_id}.fit_stats.bed.gz"
    
    @fallback_path()
    def thresholds(self):
        return self.debug_dir / f"{self.sample_id}.thresholds.bw"
    
    @fallback_path()
    def background(self):
        return self.outdir / f"{self.sample_id}.background.bw"

    def fdrs_dir(self, fdr):
        return self.outdir / f"fdr{fdr}"

    def hotspots(self, fdr):
        return str(self.fdrs_dir(fdr) / f"{self.sample_id}.hotspots.fdr{fdr}.bed.gz")
    
    def peaks(self, fdr):
        return str(self.fdrs_dir(fdr) / f"{self.sample_id}.peaks.fdr{fdr}.bed.gz")
    
    def hotspots_bb(self, fdr):
        return str(self.fdrs_dir(fdr) / f"{self.sample_id}.hotspots.fdr{fdr}.bb")
    
    def peaks_bb(self, fdr):
        return str(self.fdrs_dir(fdr) / f"{self.sample_id}.peaks.fdr{fdr}.bb")
    
    def find_missing_steps(self):
        return self.solver.find_missing_steps_from_paths()
    
    def get_display_names(self, steps):
        return self.solver.get_step_display_names(steps)


class StepsSolver:

    display_names ={
        'cutcounts': 'Extract cutcounts from BAM',
        'total_cutcounts': 'Calculate total cutcounts',
        'smoothed_signal': 'Smooth cutcounts',
        'fit_params': 'Fit background model',
        'pvals': 'Calculate p-values',
        'fdrs': 'Calculate FDRs',
        'normalized_density': 'Extract normalized density',
        'peak_calling': 'Call hotspots and peaks',
    }

    def __init__(self, paths: Hotspot3Paths, save_density):
        self.paths = paths
        self.save_density = save_density

    @staticmethod
    def resolve_required_steps(outputs, available, graph: nx.DiGraph):
        """
        Given a set of desired outputs and already available nodes,
        return the minimal set of required nodes (topologically sorted).
        """
        required = set()
        visited = set()

        def visit(node):
            if node in visited or node in available:
                return
            visited.add(node)
            for dep in graph.predecessors(node):
                visit(dep)
            required.add(node)

        for out in outputs:
            visit(out)

        return list(nx.topological_sort(graph.subgraph(required)))


    def find_missing_steps_from_paths(self):
        step_graph = nx.DiGraph()
        step_graph.add_edges_from([
            ("bam", "cutcounts"),
            ("cutcounts", "total_cutcounts"),
            
            # signal smoothing
            ("cutcounts", "smoothed_signal"),
            ("total_cutcounts", "smoothed_signal"),

            # normalized density
            ("smoothed_signal", "normalized_density"),

            # background model fitting
            ("cutcounts", "fit_params"),
            ("total_cutcounts", "fit_params"),

            # per-bp p-values
            ("fit_params", "pvals"),
            ("cutcounts", "pvals"),

            # FDR
            ("pvals", "fdrs"),

            # peak calling
            ("fdrs", "peak_calling"),
            ("smoothed_signal", "peak_calling"),
            ("total_cutcounts", "peak_calling")
        ])
        available = {x for x in step_graph.nodes if self.paths.was_set(x)}
        outputs = {'peak_calling',}
        if self.save_density:
            outputs.add('normalized_density')
        result = self.resolve_required_steps(outputs, available, step_graph)
        if 'bam' in result:
            raise ValueError("Provide a bam file or cutcounts")
        return result
    
    def get_step_display_names(self, steps):
        return [self.display_names.get(step, step) for step in steps]


import pandas as pd
import numpy as np

from hotspot3.helpers.models import FitResults, WindowedFitResults, SPOTEstimationData
from hotspot3.helpers.colors import get_bb_color
from hotspot3.helpers.stats import mean_from_r_p


bed12_columns = ['chrom', 'start', 'end', 'id', 'score', 'strand', 'thickStart', 'thickEnd', 'itemRgb', 'blockCount', 'blockSizes', 'blockStarts']


def fit_stats_df_to_fallback_fit_results(fit_result_df: pd.DataFrame):
    assert len(fit_result_df) == 1, f"Expected one fit, got {len(fit_result_df)}"
    segment_fit_params = fit_result_df[
        ['bg_p', 'bg_r', 'rmsea', 'bg_bases_prop', 'bg_tr']
    ].iloc[0].rename(
        {
            'bg_p': 'p',
            'bg_r': 'r',
            'bg_bases_prop': 'fit_quantile',
            'bg_tr': 'fit_threshold'
        }
    ).to_dict()

    return FitResults(**segment_fit_params)


def get_spot_score_fit_data(fit_data: pd.DataFrame):
    total_tags_background = fit_data['bg_tags'].values
    total_tags = fit_data['mappable_tags'].values
    total_bases = fit_data['mappable_bases'].values

    spot_scores = np.clip(1 - total_tags_background / total_tags, 0, 1)
    valid_scores = np.isfinite(spot_scores) & (spot_scores > 0) & (spot_scores < 1)
    
    return SPOTEstimationData(
        total_tags=total_tags,
        total_tags_background=total_tags_background,
        segment_spot_scores=spot_scores,
        total_bases=total_bases,
        valid_scores=valid_scores
    )

def convert_fit_results_to_series(
        fit_results: FitResults,
        fit_type: str,
        success_fit: bool,
        reached_max_q: bool=False
    ) -> pd.Series:
    return pd.Series({
        'bg_r': fit_results.r,
        'bg_p': fit_results.p,
        'rmsea': fit_results.rmsea,
        'bg_tr': fit_results.fit_threshold,
        'mappable_bases': fit_results.n_total,
        'bg_bases': fit_results.n_total - fit_results.n_signal,
        'bg_bases_prop': fit_results.fit_quantile,
        'mappable_tags': fit_results.total_tags,
        'bg_tags': fit_results.total_tags - fit_results.signal_tags,
        'fit_type': fit_type,
        'success_fit': success_fit,
        'max_bg_reached': reached_max_q
    })

def set_dtype(intervals_stats: pd.DataFrame, fit_series: pd.Series):
    """
    Workaround func to avoid future warnings about setting bool values to float (default) columns
    Initially sets value of first row to every row in the DataFrame
    """
    added_cols = ~fit_series.index.isin(intervals_stats.columns)
    if np.any(added_cols):
        for col in fit_series.index[added_cols]:
            intervals_stats[col] = fit_series[col]


def set_series_row_to_df(df: pd.DataFrame, row: pd.Series, index):
    set_dtype(df, row)
    df.loc[index, row.index] = row
    return df


def fit_results_to_df(fit_results: WindowedFitResults, per_window_trs: np.ndarray):
    return pd.DataFrame({
        'sliding_r': fit_results.r,
        'sliding_p': fit_results.p,
        'enough_bg': fit_results.enough_bg_mask,
        'tr': per_window_trs,
    })

def peaks_to_bed12(peaks_df, fdr_tr):
    """
    Convert peaks to bed9 format.
    """
    peaks_df['strand'] = '.'
    peaks_df['score'] = convert_to_score(peaks_df['max_density'], 100)
    peaks_df['thickStart'] = peaks_df['summit']
    peaks_df['thickEnd'] = peaks_df['summit'] + 1
    peaks_df['itemRgb'] = get_bb_color(fdr_tr, mode='peaks')

    peaks_df['blockCount'] = 3
    peaks_df['blockSizes'] = '1,1,1'
    peaks_df['blockStarts'] = '0,' + peaks_df.eval('summit - start').astype(str) + ',' + peaks_df.eval('end - start - 1').astype(str)

    return peaks_df[bed12_columns]


def hotspots_to_bed12(hotspots_df, fdr_tr, significant_stretches):
    """
    Convert hotspots to bed9 format.
    """
    hotspots_df['strand'] = '.'
    hotspots_df['score'] = convert_to_score(hotspots_df['max_neglog10_fdr'], 10)
    hotspots_df['thickStart'] = hotspots_df['start']
    hotspots_df['thickEnd'] = hotspots_df['end']
    hotspots_df['itemRgb'] = get_bb_color(fdr_tr, mode='hotspots')
    block_count = []
    block_sizes = []
    block_starts = []
    lengths = hotspots_df.eval('end - start').values
    for i, (starts, ends) in enumerate(significant_stretches):
        block_count.append(len(starts) + 2)

        sizes = np.pad(ends - starts, (1, 1), mode='constant', constant_values=(1, 1))
        block_sizes.append(','.join(map(str, sizes)))

        starts = np.pad(starts, (1, 1), mode='constant', constant_values=(0, lengths[i] - 1))
        block_starts.append(','.join(map(str, starts)))

    hotspots_df['blockCount'] = block_count
    hotspots_df['blockSizes'] = block_sizes
    hotspots_df['blockStarts'] = block_starts

    return hotspots_df[bed12_columns]


def convert_to_score(array, mult, max_score=1000):
    return np.round(array * mult).astype(np.int64).clip(0, max_score)
